from setuptools import setup

setup(
    name="mlneurotools",
    version="0.2.4",
    description="Tools for stats and ml",
    url="http://github.com/arthurdehgan/NeuroPy-MLToolbox",
    author="Arthur Dehgan",
    author_email="arthur.dehgan@umontreal.ca",
    packages=["mlneurotools"],
    zip_safe=False,
)
